@@ -1,13 +1,21 @@
-use crate::core::atom::Atom;
+use crate::core::atom::{Atom, is_atom_an_enum};
 use crate::core::block::Block;
 use crate::core::probe::Probe;
 use crate::core::synth::VCDValue;
 use std::collections::HashMap;
 use std::io::Write;
+use crate::core::prelude::TypeKind;
+use crate::core::type_descriptor::TypeDescriptor;
+
+#[derive(Clone, Debug)]
+enum VCDIDCode {
+    Singleton(vcd::IdCode),
+    Composite(Vec<Box<VCDIDCode>>),
+}
 
 pub struct VCDProbe<W: Write> {
     vcd: vcd::Writer<W>,
-    id_map: HashMap<usize, vcd::IdCode>,
+    id_map: HashMap<usize, VCDIDCode>,
     val_map: HashMap<vcd::IdCode, VCDValue>,
 }
 
@@ -27,6 +35,26 @@ impl<W: Write> VCDProbe<W> {
 
 struct VCDHeader<W: Write>(VCDProbe<W>);
 
+fn register_signal<W: Write>(name: &str, descriptor: &TypeDescriptor, vcd: &mut vcd::Writer<W>) -> VCDIDCode {
+    match &descriptor.kind {
+        TypeKind::Bits(width) | TypeKind::Signed(width) => {
+            VCDIDCode::Singleton(vcd.add_wire(*width as u32, name).unwrap())
+        }
+        TypeKind::Enum(_) => {
+            VCDIDCode::Singleton(vcd.add_wire(0, name).unwrap())
+        }
+        TypeKind::Composite(k) => {
+            let mut ret = vec![];
+            for field in k {
+                let sub_name = format!("{}${}", name, field.fieldname);
+                let code = register_signal(&sub_name, &field.kind, vcd);
+                ret.push(Box::new(code));
+            }
+            VCDIDCode::Composite(ret)
+        }
+    }
+}
+
 impl<W: Write> Probe for VCDHeader<W> {
     fn visit_start_scope(&mut self, name: &str, _node: &dyn Block) {
         self.0.vcd.add_module(name).unwrap();
@@ -37,13 +65,9 @@ impl<W: Write> Probe for VCDHeader<W> {
     }
 
     fn visit_atom(&mut self, name: &str, signal: &dyn Atom) {
-        let width = if signal.is_enum() {
-            0
-        } else {
-            signal.bits() as u32
-        };
-        let id = self.0.vcd.add_wire(width, name).unwrap();
-        self.0.id_map.insert(signal.id(), id);
+        self.0.id_map.insert(signal.id(),
+                             register_signal(name, &signal.descriptor(),
+                                             &mut self.0.vcd));
     }
 
     fn visit_end_namespace(&mut self, _name: &str, _node: &dyn Block) {
@@ -68,28 +92,7 @@ struct VCDChange<W: Write>(VCDProbe<W>);
 impl<W: Write> Probe for VCDChange<W> {
     fn visit_atom(&mut self, _name: &str, signal: &dyn Atom) {
         if let Some(idc) = self.0.id_map.get(&signal.id()) {
-            let val = signal.vcd();
-            if let Some(old_val) = self.0.val_map.get(idc) {
-                if val == *old_val {
-                    return;
-                }
-            }
-            self.0.val_map.insert(*idc, val.clone());
-            match val {
-                VCDValue::Single(s) => {
-                    self.0.vcd.change_scalar(*idc, s).unwrap();
-                }
-                VCDValue::Vector(v) => {
-                    if v.len() == 1 {
-                        self.0.vcd.change_scalar(*idc, v[0]).unwrap();
-                    } else {
-                        self.0.vcd.change_vector(*idc, &v).unwrap();
-                    }
-                }
-                VCDValue::String(t) => {
-                    self.0.vcd.change_string(*idc, &t).unwrap();
-                }
-            }
+            do_vcd_change(&mut self.0.val_map, &mut self.0.vcd, idc, &signal.vcd(), false);
         }
     }
 }
@@ -104,28 +107,60 @@ struct VCDDump<W: Write>(VCDProbe<W>);
 
 impl<W: Write> Probe for VCDDump<W> {
     fn visit_atom(&mut self, _name: &str, signal: &dyn Atom) {
-        if let Some(&idc) = self.0.id_map.get(&signal.id()) {
-            let val = signal.vcd();
-            self.0.val_map.insert(idc, val.clone());
+        if let Some(idc) = &self.0.id_map.get(&signal.id()) {
+            do_vcd_change(&mut self.0.val_map, &mut self.0.vcd, idc, &signal.vcd(), true);
+        }
+    }
+}
+
+fn do_vcd_change<W: Write>(val_map: &mut HashMap<vcd::IdCode, VCDValue>, vcd: &mut vcd::Writer<W>, idc: &VCDIDCode, val: &VCDValue, dump: bool) {
+    match idc {
+        VCDIDCode::Singleton(idc) => {
+            if !dump {
+                if let Some(old_val) = val_map.get(idc) {
+                    if val.eq(old_val) {
+                        return;
+                    }
+                }
+            }
+            let _ = val_map.insert(*idc, val.clone());
             match val {
                 VCDValue::Single(s) => {
-                    self.0.vcd.change_scalar(idc, s).unwrap();
+                    vcd.change_scalar(*idc, s.clone()).unwrap();
                 }
                 VCDValue::Vector(v) => {
                     if v.len() == 1 {
-                        self.0.vcd.change_scalar(idc, v[0]).unwrap();
+                        vcd.change_scalar(*idc, v[0]).unwrap();
                     } else {
-                        self.0.vcd.change_vector(idc, &v).unwrap();
+                        vcd.change_vector(*idc, &v).unwrap();
                     }
                 }
                 VCDValue::String(t) => {
-                    self.0.vcd.change_string(idc, &t).unwrap();
+                    vcd.change_string(*idc, &t).unwrap();
+                }
+                VCDValue::Composite(_) => {
+                    panic!("Composite data received for singleton type");
+                }
+            }
+        }
+        VCDIDCode::Composite(idcs) => {
+            match val {
+                VCDValue::Composite(vals) => {
+                    assert_eq!(idcs.len(), vals.len(), "Mismatch in values versus type information");
+                    for n in 0..idcs.len() {
+                        do_vcd_change(val_map, vcd, &idcs[n], &vals[n], dump);
+                    }
+                }
+                _ => {
+                    panic!("Scalar data received for composite type");
                 }
             }
         }
     }
+
 }
 
+
 pub fn write_vcd_dump<W: Write>(vcd: VCDProbe<W>, uut: &dyn Block) -> VCDProbe<W> {
     let mut visitor = VCDDump(vcd);
     visitor
